@@ -1,6 +1,6 @@
 /*                                               -*- c-file-style: "ruby"; -*-
  * mod_img - an Image Manipulatively Gadget for Apache
- * $Id: mod_img.c,v 1.1.1.1 2003/09/30 15:00:04 yuya_n Exp $
+ * $Id: mod_img.c,v 1.5 2003/10/03 07:38:15 yuya_n Exp $
  * 
  * Copyright (C) 2003  Yuya.Nishida. <yuya at@at j96 dot.dot org>
  * All rights reserved.
@@ -118,6 +118,30 @@ create_query_table(pool *pool, const char *query)
 }
 
 
+static const char *
+strerror_imlib_load_error(Imlib_Load_Error error)
+{
+    static const char *IMLIB_LOAD_ERRORS[] = {
+        "IMLIB_LOAD_ERROR_NONE",
+        "IMLIB_LOAD_ERROR_FILE_DOES_NOT_EXIST",
+        "IMLIB_LOAD_ERROR_FILE_IS_DIRECTORY",
+        "IMLIB_LOAD_ERROR_PERMISSION_DENIED_TO_READ",
+        "IMLIB_LOAD_ERROR_NO_LOADER_FOR_FILE_FORMAT",
+        "IMLIB_LOAD_ERROR_PATH_TOO_LONG",
+        "IMLIB_LOAD_ERROR_PATH_COMPONENT_NON_EXISTANT",
+        "IMLIB_LOAD_ERROR_PATH_COMPONENT_NOT_DIRECTORY",
+        "IMLIB_LOAD_ERROR_PATH_POINTS_OUTSIDE_ADDRESS_SPACE",
+        "IMLIB_LOAD_ERROR_TOO_MANY_SYMBOLIC_LINKS",
+        "IMLIB_LOAD_ERROR_OUT_OF_MEMORY",
+        "IMLIB_LOAD_ERROR_OUT_OF_FILE_DESCRIPTORS",
+        "IMLIB_LOAD_ERROR_PERMISSION_DENIED_TO_WRITE",
+        "IMLIB_LOAD_ERROR_OUT_OF_DISK_SPACE",
+        "IMLIB_LOAD_ERROR_UNKNOWN",
+    };
+
+    return IMLIB_LOAD_ERRORS[error];
+}
+
 module img_module;
 
 #define DIR_SEPARATOR "/"
@@ -134,21 +158,30 @@ typedef struct {
 typedef struct {
     const char *ext;
     const char *mime_type;
+    int save_support;
 } OutputType;
 
 const OutputType OUTPUT_TYPES[] = {
     /* {"argb", "image/x-rgb?"}, */
-    {"bmp", "image/bmp"},
-    {"gif", "image/gif"},
-    {"jpg", "image/jpeg"},
-    {"png", "image/png"},
-    {"pnm", "image/x-portable-anymap"},
+    {"bmp", "image/bmp",
+     IMLIB2_BMP_SAVE_SUPPORT},
+    {"gif", "image/gif",
+     IMLIB2_GIF_SAVE_SUPPORT},
+    {"jpg", "image/jpeg",
+     IMLIB2_JPG_SAVE_SUPPORT},
+    {"png", "image/png",
+     IMLIB2_PNG_SAVE_SUPPORT},
+    {"pnm", "image/x-portable-anymap",
+     IMLIB2_PNM_SAVE_SUPPORT},
     /* {"tga", "?"}, */
-    {"tiff", "image/tiff"},
-    {"xpm", "image/x-pixmap"},
+    {"tiff", "image/tiff",
+     IMLIB2_TIFF_SAVE_SUPPORT},
+    {"xpm", "image/x-pixmap",
+     IMLIB2_XPM_SAVE_SUPPORT},
     {NULL}
 };
-const OutputType DEFAULT_OUTPUT_TYPE = {"jpg", "image/jpeg"};
+const OutputType DEFAULT_OUTPUT_TYPE = {"jpg", "image/jpeg",
+                                        IMLIB2_JPG_SAVE_SUPPORT};
 
 static const OutputType *ext2output_type(const char *ext)
 {
@@ -248,10 +281,18 @@ manipulate_file_mimetype(ManipulaterData *md)
 {
     const OutputType *output_type = mime_type2output_type(md->arg_value);
 
-    if (output_type != NULL) {
+    if (output_type == NULL) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, md->r, 
+                      "Unknown mimetype. (%s)", 
+                      md->arg_value); 
+        return HTTP_BAD_REQUEST;
+    } else {
         md->output_type = output_type;
     }
 
+    imlib_context_set_image(md->ii_in);
+    md->ii_out = imlib_clone_image();
+
     return OK;
 }
 
@@ -400,6 +441,7 @@ static void
 img_initialize(server_rec *s, pool *p)
 {
     ap_add_version_component(MOD_IMG_VERSION);
+    ap_add_version_component("Imlib2/" IMLIB2_VERSION);
 }
 
 static void *
@@ -550,6 +592,17 @@ get_merged_config(request_rec *r)
     cfg_default.max_ratio = cfg->max_ratio;
     cfg = img_merge_config(r->pool, &cfg_default, cfg);
 
+    if (cfg->output_type == &DEFAULT_OUTPUT_TYPE) {
+        request_rec *subr;
+        const OutputType *output_type;
+
+        subr = ap_sub_req_lookup_file(r->filename, r);
+        output_type = mime_type2output_type(subr->content_type);
+        if (output_type != NULL) {
+            cfg->output_type = output_type;
+        }
+    }
+
     return (ImgConfig *)cfg;
 }
 
@@ -608,11 +661,21 @@ img_handler(request_rec *r)
         char *workdir;
         const char *workfilename;
         CleanupRmRecursiveData *crrd;
+        Imlib_Load_Error error_return;
 
         md.r = r;
         md.cfg = cfg;
         md.output_type = cfg->output_type;
-        md.ii_out = md.ii_in = imlib_load_image(r->filename);
+        md.ii_out = md.ii_in =
+            imlib_load_image_with_error_return(r->filename, &error_return);
+
+        if (error_return != IMLIB_LOAD_ERROR_NONE) {
+            ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
+                          "img_handler: load error (r->filename=%s, error_return=%s)",
+                          r->filename,
+                          strerror_imlib_load_error(error_return));
+            return HTTP_INTERNAL_SERVER_ERROR;
+        }
 
         workdir = ap_pstrcat(r->pool,
                              cfg->workdir, DIR_SEPARATOR,
@@ -652,6 +715,11 @@ img_handler(request_rec *r)
             md.ii_in = md.ii_out;
         }
 
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
+                      "img_handler: creating workfile (filename=%s, args=%s, workdir=%s, mime_type=%s, ext=%s)",
+                      r->filename, r->args, workdir,
+                      md.output_type->mime_type, md.output_type->ext);
+
         r->content_type = md.output_type->mime_type;
         workfilename = ap_pstrcat(r->pool,
                                   workdir, DIR_SEPARATOR,
@@ -659,8 +727,30 @@ img_handler(request_rec *r)
 
         imlib_context_set_image(md.ii_out);
         imlib_image_set_format(md.output_type->ext);
-        imlib_save_image(workfilename);
+        imlib_save_image_with_error_return(workfilename,
+                                           &error_return);
         imlib_free_image();
+        if (error_return != IMLIB_LOAD_ERROR_NONE) {
+            int result = HTTP_INTERNAL_SERVER_ERROR;
+
+            if (!md.output_type->save_support) {
+                ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
+                              "Converting to `%s' is not supported.",
+                              md.output_type->mime_type);
+                ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
+                              "img_handler: Imlib2/" IMLIB2_VERSION
+                              " does not support `%s' saving.",
+                              md.output_type->ext);
+                result = HTTP_NOT_IMPLEMENTED;
+            }
+
+            ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
+                          "img_handler: save error (workfilename=%s, error_return=%s)",
+                          workfilename,
+                          strerror_imlib_load_error(error_return));
+
+            return result;
+        }
 
         ap_pfclose(r->pool, f);
         f = ap_pfopen(r->pool, workfilename, "rb");
